# mfrc_read_sector.py

# liest
# - die UID des Tags (RFID-Karte/Chip) und gibt sie auf dem Display und dem Terminal aus
# - die Blöcke eines Sektors und gibt sie inkl. ACCESS-Bits auf dem Terminal aus

# Quelle: https://github.com/Tasm-Devil/micropython-mfrc522-esp32; ergänzt und modifiziert für TTGO-T-Display

# Achtung: Vor dem Start des Programms muss das Modul mfrc522.py auf den ESP32 hochgeladen werden.


# Initialierungen...

from machine import Pin, SPI, SoftSPI
import st7789
import vga2_16x16 as font1 # mit Umlauten
import vga2_8x16 as font2
from os import uname
release = uname().release[0:4]
id = 2 # z. B. für V 1.12 und V 1.14
if release == '1.18':
    id = 1    
spi = SPI(id, baudrate=20000000, polarity=1, sck=Pin(18), mosi=Pin(19)) # 1. Parameter = 1 für Firmware v1.18
display = st7789.ST7789(spi, 135, 240,  reset=Pin(23, Pin.OUT), cs=Pin(5, Pin.OUT), dc=Pin(16, Pin.OUT), backlight=Pin(4, Pin.OUT), rotation=3)
display.init()

from time import sleep_ms
from mfrc522 import MFRC522

sck = Pin(17, Pin.OUT)
mosi = Pin(15, Pin.OUT)
miso = Pin(12, Pin.OUT)
softspi = SoftSPI(baudrate=100000, polarity=0, phase=0, sck=sck, mosi=mosi, miso=miso)

sda = Pin(25, Pin.OUT) # (SDA = ChipSelect)


# Funktionen...

def umlaute_ersetzen(zk):
    zk = zk.replace(b'Ä', b'\x8e')  
    zk = zk.replace(b'ä', b'\x84')
    zk = zk.replace(b'Ö', b'\x99')
    zk = zk.replace(b'ö', b'\x94')
    zk = zk.replace(b'Ü', b'\x9a')
    zk = zk.replace(b'ü', b'\x81')
    zk = zk.replace(b'ß', b'\xe1')
    return zk

def show_as_string(b, l):
    print('Block['+str(b) + ']: ', end = '')
    for i in l:
        if i >= 32 and i <= 126:
            print(chr(i), end = '')
        else:
            print('.', end = '')
    print()
    
def bitliste(n):
    return [1 if n & (1 << (7-i)) else 0 for i in range(8)]

def do_access_bits(access_bytes):    
    access_bits = [[0,0,0] for i in range(4)] # 4 mal [C1, C2, C3]
    sechs = 6
    sieben = 7
    for i in range(4):
        access_bits[i] = [bitliste(access_bytes[7-sechs])[sieben-(4+i)], bitliste(access_bytes[8-sechs])[sieben-i], bitliste(access_bytes[8-sechs])[sieben-(4+i)]]    
    return access_bits    

def do_read():
    try:
        while True:
            # print()
            rdr = MFRC522(softspi, sda)
            uid = ""
            (stat, tag_type) = rdr.request(rdr.REQIDL) # liefert stat (stat = 2 > Err; stat = 0 > OK)
            # print('stat=', stat, ' tag_type=', tag_type) # zum Testen (tag_type = 16 > 
            if stat != rdr.OK: # Karte/Chip nicht vorhanden
                print('Karte platzieren!')
            else:          
                (stat, raw_uid) = rdr.anticoll()
                # rdr.anticoll() kontrolliert, ob nur ein einziges RFID-Objekt detektiert wird (d. h. keine Collision vorliegt);
                # in diesem Fall erhält stat den Wert rdr.OK und raw_uid eine Liste mit 4 Bytes
                
                # UID lesen und anzeigen:
                if stat == rdr.OK: 
                    uid = ('0x%02x%02x%02x%02x' % (raw_uid[0], raw_uid[1], raw_uid[2], raw_uid[3])) # setzt 4 Bytes zu einem "HEX-String" zusammen
                    # print('raw_uid: ', raw_uid) # dezimal
                    # print('raw_uid: ', [hex(n) for n in raw_uid]) # hexadezimal
                    print('uid: ', uid)
                    print()
                    display.text(font1, uid, 5, 50)
                    sleep_ms(100)
                
                # einen Sektor lesen und anzeigen
                sector = int(input('Sektornummer (0 - 15): '))
                if sector > 15:
                    print('Sektornummer zu groß... Abbruch!')
                    break
                if rdr.select_tag(raw_uid) == rdr.OK:
                    key = [0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF] # für Authentifizierung
                    block0 = sector * 4
                    for block in range(block0, block0+4):
                      if rdr.auth(rdr.AUTHENT1A, block, key, raw_uid) == rdr.OK:
                        newdata = rdr.read(block)
                        # print(newdata)
                        # print('Block['+str(block) + ']: ', newdata) # Dezimal
                        print('Block['+str(hex(block)) + ']: ', [hex(n) for n in newdata]) # Hexadezimal
                        # show_as_string(block, newdata) # Zeichenkette
                        if block % 4 == 3:
                            access_bytes = newdata[6:9]
                            access_bits = do_access_bits(access_bytes)
                            print('Access-Bits: ', access_bits)
                            print()
                      else:
                        print('Authentifizierungsfehler')
                        rdr.stop_crypto1()                        
                else:
                    print('Tag nicht vorhanden')    
                print()
                print()
                ende = input('noch einmal? j/n ')
                if ende == 'n':
                    print('Tschüss')
                    break
                
    except KeyboardInterrupt:
        print()
        print()
        print("Tschüss")
        display.text(font1, umlaute_ersetzen(b'Tschüss!     '), 5, 90)
        
        
# Hauptprogramm...

display.fill(0)
display.text(font1, 'RFID-RC522', 5, 10)
display.text(font1, 'Karte!', 5, 50)
display.text(font2, 'Exit mit Strg-C', 5, 90)
do_read() 